/*
 * ml3.c
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995-97)
 * this file belongs to the PUZZLE version 3.0 distribution
 *
 */


#define EXTERN extern


/* prototypes */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "util.h"
#include "ml.h"


/******************************************************************************/
/* chi square test (Numerical Recipes)                                        */
/******************************************************************************/

/* start code of Press et al. "Numerical Recipes in C" (1992) */

double gammln(double xx)
{
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
		24.01409824083091,-1.231739572450155,
		0.1208650973866179e-2,-0.5395239384953e-5};
	int j;

	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++) ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}

#define ITMAX 100
#define EPS 3.0e-7
#define FPMIN 1.0e-30

void gcf(double *gammcf, double a, double x, double *gln)
{
	int i;
	double an,b,c,d,del,h;

	*gln=gammln(a);
	b=x+1.0-a;
	c=1.0/FPMIN;
	d=1.0/b;
	h=d;
	for (i=1;i<=ITMAX;i++) {
		an = -i*(i-a);
		b += 2.0;
		d=an*d+b;
		if (fabs(d) < FPMIN) d=FPMIN;
		c=b+an/c;
		if (fabs(c) < FPMIN) c=FPMIN;
		d=1.0/d;
		del=d*c;
		h *= del;
		if (fabs(del-1.0) < EPS) break;
	}	
	*gammcf=exp(-x+a*log(x)-(*gln))*h;
}
#undef ITMAX
#undef EPS
#undef FPMIN

#define ITMAX 100
#define EPS 3.0e-7

void gser(double *gamser, double a, double x, double *gln)
{
	int n;
	double sum,del,ap;

	*gln=gammln(a);
	if (x <= 0.0) {
		*gamser=0.0;
		return;
	} else {
		ap=a;
		del=sum=1.0/a;
		for (n=1;n<=ITMAX;n++) {
			++ap;
			del *= x/ap;
			sum += del;
			if (fabs(del) < fabs(sum)*EPS) {
				*gamser=sum*exp(-x+a*log(x)-(*gln));
				return;
			}
		}
		*gamser=sum*exp(-x+a*log(x)-(*gln));
		return;
	}
}
#undef ITMAX
#undef EPS

double gammq(double a, double x)
{
	double gamser,gammcf,gln;

	if (x < (a+1.0)) {
		gser(&gamser,a,x,&gln);
		return 1.0-gamser;
	} else {
		gcf(&gammcf,a,x,&gln);
		return gammcf;
	}
}

double gammp(double a, double x)
{
	double gamser,gammcf,gln;

	if (x < (a+1.0)) {
		gser(&gamser,a,x,&gln);
		return gamser;
	} else {
		gcf(&gammcf,a,x,&gln);
		return 1.0-gammcf;
	}
}

/* error function */
double myerff(double x)
{
	return (x < 0.0 ? -gammp(0.5,x*x) : gammp(0.5,x*x));
}

/* end of NR code */

/* chi square test
   ef      expected frequencies (sum up to 1 !!)
   of      observed frequencies (sum up to the number of samples)
   numcat  number of categories
   returns critical significance level */
double chi2test(double *ef, int *of, int numcat)
{	
	double chi2, criticals, efn;
	int i, samples, below1, below5;
	
	chi2fail = FALSE;
	below1 = 0;
	below5 = 0;
	
	/* compute number of samples */
	samples = 0;
	for (i = 0; i < numcat; i++)
		samples = samples + of[i];
	
	/* compute chi square */
	chi2 = 0;
	for (i = 0; i < numcat; i++) {
		efn = ef[i]*((double) samples);
		if (efn < 1.0) below1++;
		if (efn < 5.0) below5++;
		chi2 = chi2 + ((double) of[i]-efn)*((double) of[i]-efn)/efn;
	}

	/* compute significance if chi2 were the critical value */
	criticals = gammq((numcat - 1)*0.5, chi2*0.5);
	
	/* no expected frequency category (sum up to # samples) below 1.0 */
	if (below1 > 0) chi2fail = TRUE;
	/* no more than 1/5 of the frequency categories below 5.0 */
	if (below5 > (int) floor(samples/5.0)) chi2fail = TRUE;

	return criticals;
}

/* compare general base frequencies with frequencies of taxon i with chi square */
double homogentest(int taxon)
{
	return chi2test(Freqtpm, Basecomp[taxon], gettpmradix());
}


/******************************************************************************/
/* discrete Gamma-distribution                                                */
/******************************************************************************/

/* Algorithms for the inverse of regularized incomplete gamma function and
   for the inverse error function are adopted from the Mathematica package
   "InverseStatisticalFunctions.m" by Jerry B. Keiper (version 1991).
   This package can be freely obtained from http://www.wri.com/mathsource
   and is copyright 1989-1993, Wolfram Research, Inc. */

/* inverse of error function */
double inverf(double pp) {

	double x, x0, p, erfxx, z;
	int cflag, rflag, i;
		
	x = 0.0;
	cflag = FALSE;
	p = pp;
		
	if (p == 0.0) return 0.0;

	if (p < 0) rflag = TRUE;
	else rflag = FALSE;	
	if (rflag) p = -p;

	if (p < 0.9) {
	
		x0 = (0.8862244286 + (-1.220051315 +
				(0.2194688168 + 0.1329123619*p)*p)*p)*p/
				(1.0 - (1.37690286 + (0.01110212605 +
				(-0.4939796583 + 0.0964477892*p)*p)*p)*p);
		if (x0 > 0.5) cflag = TRUE;
		else cflag = FALSE;
		if (cflag) p = 1.0 - p;
			
	} else {
		    
		if (cflag) {
			if (rflag) p = pp;
			else p = -pp;
		} else {
			p = 1.0 - p;
		}
		
		cflag = TRUE;
			
		if (p <= 0.0) {
		printf("\n\n\nHALT: PLEASE REPORT ERROR N TO DEVELOPERS\n\n\n");
		exit(1);
		}
		
		x0 = sqrt( (-0.555075007823429663 + p*(-22943097299079.457 +
				p*(-3.90911954269625125e21 + p*(-4.20106086583717546e26 +
				p*(-2.42091291012776922e29 + p*(7.8542986097983114e29))))))/
				(1 + p*(41894651750846.3689 + p*(7.23863012895805264e21 +
				p*(7.9066722226489883e26 + p*(4.63843915091736701e29 +
				p*(-1.56919477544184394e30))))))-log(p)-log(-log(p))*0.5);
	}

	
	erfxx = p;
	p = 0.886226925452758;
		
	i = 0;
		
	while (i <15) {
		
		if (i > 0) {
			if (x == 0.0)
				i = 3*i;
			else {
				z = (int) floor(-8.0*log(fabs(x)) + 0.5);
				z = (z > 1 ? z : 1);
				i = (3*i < z ? 3*i : z);
			}
			if (i > 15) i = (i > 30 ? 15 : 14);
			 
		} else {
			i = 1;
		}
			
		if (cflag)
			x = (erfxx - 1.0 + myerff(x0))*p*exp(x0*x0);
		else
			x = (myerff(x0) - erfxx)*p*exp(x0*x0);

		x *= 1.0 - x*x0/(1.0 + 2*x*x0);
		x0 -= x;
	}

	if (rflag) return -x0;
	else return x0;
}

/* inverse of regularized incomplete gamma function
    restrictions: a > 0  &&  x >= 0  &&  1 > pp >= 0 */
double invgamreg(double a, double x, double pp) {
#define EPS 0.00001

	int j, cflag;
	double gamx, p, gama, x0, x1, x2, temp, pa, dy, z;
	 
	gama = exp(gammln(a));
	
	if (x > a) cflag = TRUE;
	else cflag = FALSE;

	if (cflag) {
		gamx = gammq(a,x);
		p = pp + 1.0 - gamx;
	} else {
		gamx = (gammq(a,0.0)-gammq(a,x));
		p = pp + gamx;
	}

	/* p == GammaRegularized[a,0,x0] where x0 is still unknown
		gamx is the value of the incomplete gamma function from
		0 to x or x to Infinity depending on cflag. */

	if (p == 0.0) return 0.0;
	
	if (p < 0.0 || p > 1.0) {
		printf("\n\n\nHALT: PLEASE REPORT ERROR L TO DEVELOPERS\n\n\n");
		exit(1);
	}
	
	if(a == 1.0) {
		if (cflag) p = gamx - pp;
		else p = 1.0 - p;
		return (-log(p));
	}

	pa = pow((p*a*gama),1.0/a);
	temp = pa/(a+1.0);

	x0 = pa*(1.0+temp*(1.0+temp*((5.0+3.0*a)/(2.0*(a+2.0))+temp*((31.0+a*
		(33.0+8.0*a))/(3.0*(2.0+a)*(3.0+a))+temp*((2888.0+a*(5661.0+a*
		(3971.0+a*(1179.0+125.0*a))))/(24.0*(2.0+a)*(2.0+a)*(3.0+a)*(4.0+a))+
		temp*((9074.0+a*(20997.0+a*(18375.0+a*(7575.0+a*(1471.0+108.0*a)))))/
		(10.0*(2.0+a)*(2.0+a)*(3.0+a)*(4.0+a)*(5.0+a))))))));

	if (x0 > 5.0) { /* if x0 is large it may be very wrong */
		if (cflag)
			temp = inverf(1.0 + 2.0*(pp - gamx));
		else
			temp = inverf(2.0*p - 1.0);
		temp = 1.0 - 1.0/(9.0*a) + temp/sqrt(4.5*a);
		x1 = a*temp*temp*temp;
		/* if x1 is small it may be very wrong */
		if(x1 > 5.0) x0 = x1;
	}

	if(x0 > a) {
		if (cflag)
			p = gamx - pp;
		else {
			cflag = TRUE;
			p = 1.0 - p;
		}
	} else
		cflag = FALSE;
	
	if (p < 0.0) {
		printf("\n\n\nHALT: PLEASE REPORT ERROR M TO DEVELOPERS\n\n\n");
		exit(1);
	}

	/* now use Taylor series to quartically converge to x0 */
	p *= gama;
	j = 0;
	while (j < 15) { /* 15 digits accuracy */

		if (j > 0) {
			if(x1 == 0.0) {
				j = 4*j;
			} else {
				z = (int) floor(-1.7*log(fabs(x1)) + 0.5);
				z = (z < 1 ? 1 : z);
				z = (4*j < z ? 4*j : z);
				j = (j > z ? j : z);
			}
			if(j > 15) {
				if (j > 30) j = 15;
				else j = 14;
			}	
		} else {
			j = 1;
		}
		
		if (cflag)
			dy = p-gama*gammq(a,x0);
		else
			dy = p-gama*(gammq(a,0.0) - gammq(a,x0));

		while (fabs(dy) > EPS && j < 15) {
			/* adding a fake zero would corrupt everything */
			j = 15;
			if (cflag)
				dy = p-gama*gammq(a,x0);
			else
				dy = p-gama*(gammq(a,0.0) - gammq(a,x0));
		}

		if (fabs(dy) <= EPS) return x0;	/* success */
		
		if (cflag) dy = -dy;

		x1 = exp(x0)*dy*pow(x0,-a);
		x2 = x1*((1.0-a+x0)/2.0+x1*((1.0+a*(-3.0+2.0*a)+(4.0-4.0*a+2.0*x0)*x0)/
			6.0+x1*(1.0+a*(-6.0+a*(11.0-6.0*a))+x0*(11.0+a*(-29.0+18.0*a)+
			x0*(18.0-18.0*a+6.0*x0)))/24.0));
		x1 *= x0*(1.0+x2);
		x0 += x1;
	}
	return x0;
}
#undef EPS

/* compute rates of each category when rates are Gamma-distributed */
void updaterates()
{
	int i;
	double mean, alpha;
	
	if (numcats == 1) {
		Rates[0] = 1.0;
		return;
	}
	if (Geta == 0.0) {
		for (i = 0; i < numcats; i++)
			Rates[i] = 1.0;
		return;
	}
	alpha = (1.0 - Geta)/Geta;
	Rates[0] = invgamreg(alpha,0.0,1.0/(2.0*numcats));
	for (i = 1; i < numcats; i++) {
		Rates[i] = invgamreg(alpha,0.0,(2.0*i+1.0)/(2.0*numcats));
		/* in case that invgamreg did not converge */
		if (Rates[i] < Rates[i-1])
			Rates[i] = Rates[i-1];
	}
	/* make mean of discrete distribution equal to one */
	mean = 0.0;
	for (i = 0; i < numcats; i++) mean += Rates[i];
	mean = mean/(double)numcats;
	for (i = 0; i < numcats; i++) Rates[i] = Rates[i]/mean;
	/* check for very small rates */
	for (i = 0; i < numcats; i++)
		if (Rates[i] < 0.000001) Rates[i] = 0.000001;
}

/******************************************************************************/
/* minimization of a function by Brents method (Numerical Recipes)            */
/******************************************************************************/

#define ITMAX 100
#define CGOLD 0.3819660
#define GOLD 1.618034
#define GLIMIT 100.0
#define TINY 1.0e-20
#define ZEPS 1.0e-10
#define SHFT(a,b,c,d) (a)=(b);(b)=(c);(c)=(d);
#define SIGN(a,b) ((b) >= 0.0 ? fabs(a) : -fabs(a))

/* Brents method in one dimension */
double brent(double ax, double bx, double cx, double (*f)(double), double tol,
	double *foptx, double *f2optx, double fax, double fbx, double fcx)
{
	int iter;
	double a,b,d,etemp,fu,fv,fw,fx,p,q,r,tol1,tol2,u,v,w,x,xm;
	double xw,wv,vx;
	double e=0.0;

	a=(ax < cx ? ax : cx);
	b=(ax > cx ? ax : cx);
	x=bx;
	fx=fbx;
	if (fax < fcx) {
		w=ax;
		fw=fax;
		v=cx;
		fv=fcx;
	} else {
		w=cx;
		fw=fcx;
		v=ax;
		fv=fax;	
	}
	for (iter=1;iter<=ITMAX;iter++) {
		xm=0.5*(a+b);
		tol2=2.0*(tol1=tol*fabs(x)+ZEPS);
		if (fabs(x-xm) <= (tol2-0.5*(b-a))) {
			*foptx = fx;
			xw = x-w;
			wv = w-v;
			vx = v-x;
			*f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
				(v*v*xw + x*x*wv + w*w*vx);
			return x;
		}
		if (fabs(e) > tol1) {
			r=(x-w)*(fx-fv);
			q=(x-v)*(fx-fw);
			p=(x-v)*q-(x-w)*r;
			q=2.0*(q-r);
			if (q > 0.0) p = -p;
			q=fabs(q);
			etemp=e;
			e=d;
			if (fabs(p) >= fabs(0.5*q*etemp) || p <= q*(a-x) || p >= q*(b-x))
				d=CGOLD*(e=(x >= xm ? a-x : b-x));
			else {
				d=p/q;
				u=x+d;
				if (u-a < tol2 || b-u < tol2)
					d=SIGN(tol1,xm-x);
			}
		} else {
			d=CGOLD*(e=(x >= xm ? a-x : b-x));
		}
		u=(fabs(d) >= tol1 ? x+d : x+SIGN(tol1,d));
		fu=(*f)(u);
		if (fu <= fx) {
			if (u >= x) a=x; else b=x;
			SHFT(v,w,x,u)
			SHFT(fv,fw,fx,fu)
		} else {
			if (u < x) a=u; else b=u;
			if (fu <= fw || w == x) {
				v=w;
				w=u;
				fv=fw;
				fw=fu;
			} else if (fu <= fv || v == x || v == w) {
				v=u;
				fv=fu;
			}
		}
	}
	*foptx = fx;
	xw = x-w;
	wv = w-v;
	vx = v-x;
	*f2optx = 2.0*(fv*xw + fx*wv + fw*vx)/
		(v*v*xw + x*x*wv + w*w*vx);
	return x;
}
#undef ITMAX
#undef CGOLD
#undef ZEPS
#undef SHFT
#undef SIGN
#undef GOLD
#undef GLIMIT
#undef TINY

/* one-dimensional minimization - as input a lower and an upper limit and a trial
  value for the minimum is needed: xmin < xguess < xmax
  the function and a fractional tolerance has to be specified
  onedimenmin returns the optimal x value and the value of the function
  and its second derivative at this point
  */
double onedimenmin(double xmin, double xguess, double xmax, double (*f)(double),
	double tol, double *fx, double *f2x)
{
	double eps, optx, ax, bx, cx, fa, fb, fc;
		
	/* first attempt to bracketize minimum */
	eps = xguess*tol*50.0;
	ax = xguess - eps;
	if (ax < xmin) ax = xmin;
	bx = xguess;
	cx = xguess + eps;
	if (cx > xmax) cx = xmax;
	
	/* check if this works */
	fa = (*f)(ax);
	fb = (*f)(bx);
	fc = (*f)(cx);

	/* if it works use these borders else be conservative */
	if ((fa < fb) || (fc < fb)) {
		if (ax != xmin) fa = (*f)(xmin);
		if (cx != xmax) fc = (*f)(xmax);
		optx = brent(xmin, xguess, xmax, f, tol, fx, f2x, fa, fb, fc);
	} else
		optx = brent(ax, bx, cx, f, tol, fx, f2x, fa, fb, fc);

	return optx; /* return optimal x */
}

/* two-dimensional minimization with borders and calculations of standard errors */
/* we optimize along basis vectors - not very optimal but it seems to work well */
void twodimenmin(double tol,
	int active1, double min1, double *x1, double max1, double (*func1)(double), double *err1,
	int active2, double min2, double *x2, double max2, double (*func2)(double), double *err2)
{
	int it, nump, change;
	double x1old, x2old;
	double fx, f2x;

	it = 0;
	nump = 0;

	/* count number of parameters */
	if (active1) nump++;
	if (active2) nump++;

	do { /* repeat until nothing changes any more */
		it++;
		change = FALSE;

		/* optimize first variable */
		if (active1) {

			if ((*x1) <= min1) (*x1) = min1 + 0.2*(max1-min1);
			if ((*x1) >= max1) (*x1) = max1 - 0.2*(max1-min1);
			x1old = (*x1);
			(*x1) = onedimenmin(min1, (*x1), max1, func1, tol, &fx, &f2x);
			if ((*x1) < min1) (*x1) = min1;
			if ((*x1) > max1) (*x1) = max1;
			/* same tolerance as 1D minimization */
			if (fabs((*x1) - x1old) > 3.3*tol) change = TRUE;
			
			/* standard error */
			f2x = fabs(f2x);
			if (1.0/(max1*max1) < f2x) (*err1) = sqrt(1.0/f2x);
			else (*err1) = max1;

		}

		/* optimize second variable */
		if (active2) {

			if ((*x2) <= min2) (*x2) = min2 + 0.2*(max2-min2);
			if ((*x2) >= max2) (*x2) = max2 - 0.2*(max2-min2);
			x2old = (*x2);
			(*x2) = onedimenmin(min2, (*x2), max2, func2, tol, &fx, &f2x);
			if ((*x2) < min2) (*x2) = min2;
			if ((*x2) > max2) (*x2) = max2;
			/* same tolerance as 1D minimization */
			if (fabs((*x2) - x2old) > 3.3*tol) change = TRUE;
			
			/* standard error */
			f2x = fabs(f2x);
			if (1.0/(max2*max2) < f2x) (*err2) = sqrt(1.0/f2x);
			else (*err2) = max2;

		}

		if (nump == 1) return;
		
	} while (it != MAXITS && change);

	return;
}


/******************************************************************************/
/* parameter estimation                                                       */
/******************************************************************************/

/* compute sample mean and standard deviation of sample mean */
void computestat(double *data, int n, double *mean, double *err)
{	
	int i;
	double sum;
	
	sum = 0;
	for (i = 0; i < n; i++) sum += data[i];
	(*mean) = sum/(double) n;
	
	sum = 0;
	for (i = 0; i < n; i++) sum += (data[i] - (*mean))*(data[i] - (*mean));
	if (n != 1)
		(*err) = sqrt(sum/((double) n*(n-1))); /* unbiased estimator */
	else
		(*err) = 0.0; /* if n == 1 */
}

/* compute ML value of quartet (a,b,c,d) */
double quartetml(int a, int b, int c, int d)
{
	double d1, d2, d3;

	/* compute ML for all topologies */
	if (approxp_optn) { /* approximate parameter mode */
		d1 = quartet_alklhd(a,b,c,d); /* (a,b)-(c,d) */
		d2 = quartet_alklhd(a,c,b,d); /* (a,c)-(b,d) */
		d3 = quartet_alklhd(a,d,b,c); /* (a,d)-(b,c) */
	} else {
		d1 = quartet_lklhd(a,b,c,d); /* (a,b)-(c,d) */
		d2 = quartet_lklhd(a,c,b,d); /* (a,c)-(b,d) */
		d3 = quartet_lklhd(a,d,b,c); /* (a,d)-(b,c) */
	}
	
	/* looking for max(d1, d2, d3) */						
	if (d1 < d2) { /* d2 > d1 */
		if (d2 < d3) { /* d3 > d2 > d1 */
			/* d3 maximum */ 
			return d3;
		} else { /* d2 >= d3 > d1 */
			/* d2 maximum */
			return d2;
		}
	} else { /* d1 >= d2 */
		if (d1 < d3) { /* d3 > d1 >= d2 */
			/* d3 maximum */
			return d3;
		} else { /* d1 >= d2 && d1 >= d3 */
			/* d1 maximum */
			return d1;
		}
	}
}

/* optimization function TSparam - quartets */
double opttsq(double x)
{
	if (x < MINTS) TSparam = MINTS;
	else if (x > MAXTS) TSparam = MAXTS;
	else TSparam = x;
	tranprobmat();
	distupdate(qca, qcb, qcc, qcd);	
	return (-quartetml(qca, qcb, qcc, qcd));
}

/* optimization function YRparam - quartets */
double optyrq(double x)
{
	if (x < MINYR) YRparam = MINYR;
	else if (x > MAXYR) YRparam = MAXYR;
	else YRparam = x;
	tranprobmat();
	distupdate(qca, qcb, qcc, qcd);
	return (-quartetml(qca, qcb, qcc, qcd));
}

/* estimate substitution process parameters - random quartets */
void optimseqevolparamsq()
{
	double tsmeanold, yrmeanold;
	dvector tslist, yrlist;
	int fin;
	ivector taxon;
	uli minqts, maxqts, n;


	taxon = new_ivector(4);

	/* number of quartets to be investigated */
	minqts = (uli) floor(0.25 * MINPERTAXUM * Maxspc) + 1;
	maxqts = (uli) floor(0.25 * MAXPERTAXUM * Maxspc) + 1;
	if (Maxspc == 4) {
		minqts = (uli) 1;
		maxqts = (uli) 1;
	}
	
	tslist = new_dvector(maxqts);
	yrlist = new_dvector(maxqts);
	
	/* initialize averages */
	tsmean = TSparam;
	yrmean = YRparam;

	fin = FALSE;
		
	/* investigate maxqts random quartets */
	for (n = 0; n < maxqts; n++) {
	
		/* choose random quartet */
		chooser(Maxspc, 4, taxon);

		/*
		 * optimize parameters on this quartet
		 */

		qca = taxon[0];
		qcb = taxon[1];
		qcc = taxon[2];
		qcd = taxon[3];
		
		/* initialize start values with average value */
		if ((SH_optn || nuc_optn) && optim_optn && (data_optn == 0)) TSparam = tsmean;
		if ((nuc_optn && TN_optn) && optim_optn && (data_optn == 0)) YRparam = yrmean;
		
		/* estimation */
		twodimenmin(PEPS1,
			(SH_optn || nuc_optn) && optim_optn && (data_optn == 0),
				MINTS, &TSparam, MAXTS, opttsq, &tserr,
			(nuc_optn && TN_optn) && optim_optn && (data_optn == 0),
				MINYR, &YRparam, MAXYR, optyrq, &yrerr);


		tsmeanold = tsmean;
		yrmeanold = yrmean;
		tslist[n] = TSparam;
		yrlist[n] = YRparam;
		computestat(tslist, n+1 , &tsmean, &tserr);
		computestat(yrlist, n+1 , &yrmean, &yrerr);

		/* check whether the means are converging */
		if (n > minqts-2) {
			if ((fabs(tsmean-tsmeanold) < TSDIFF) &&
				(fabs(yrmean-yrmeanold) < YRDIFF))
					fin = TRUE;
		}

		/* investigate at least minqts quartets */
		if (n > minqts-2 && (fin || n > maxqts-2)) break;
	}

	/* round estimated numbers to 2 digits after the decimal point */
	if (tserr != 0.0) tsmean = floor(100.0*tsmean+0.5)/100.0;
	if (yrerr != 0.0) yrmean = floor(100.0*yrmean+0.5)/100.0;

	/* update ML engine */
	TSparam = tsmean;
	YRparam = yrmean;
	tranprobmat();

	free_ivector(taxon);
}

/* optimization function TSparam - tree */
double opttst(double x)
{
	double result;

	if (x < MINTS) TSparam = MINTS;
	else if (x > MAXTS) TSparam = MAXTS;
	else TSparam = x;
	tranprobmat();
	computedistan();
	if (approxp_optn) result = usertree_alklhd();
	else result = usertree_lklhd();

	return (-result);
}

/* optimization function YRparam - tree */
double optyrt(double x)
{
	double result;
	
	if (x < MINYR) YRparam = MINYR;
	else if (x > MAXYR) YRparam = MAXYR;
	else YRparam = x;
	tranprobmat();
	computedistan();
	if (approxp_optn) result = usertree_alklhd();
	else result = usertree_lklhd();

	return (-result);
}


/* optimize substitution process parameters - tree */
void optimseqevolparamst()
{
	twodimenmin(PEPS1,
		(SH_optn || nuc_optn) && optim_optn && (data_optn == 0),
			MINTS, &TSparam, MAXTS, opttst, &tserr,
		(nuc_optn && TN_optn) && optim_optn && (data_optn == 0),
			MINYR, &YRparam, MAXYR, optyrt, &yrerr);
}


/* optimization function fracinv */
double optfi(double x)
{
	double result;

	if (x < MINFI) fracinv = MINFI;
	else if (x > fracconst) fracinv = fracconst;
	else fracinv = x;
	
	computedistan();
	if (approxp_optn) result = usertree_alklhd();
	else result = usertree_lklhd();

	return (-result);	
}


/* optimization function Geta */
double optge(double x)
{
	double result;

	if (x < MINGE) Geta = MINGE;
	else if (x > MAXGE) Geta = MAXGE;
	else Geta = x;
	
	updaterates();
	
	computedistan();
	if (approxp_optn) result = usertree_alklhd();
	else result = usertree_lklhd();

	return (-result);	
}


/* optimize rate heterogenity parameters */
void optimrateparams()
{	
	twodimenmin(PEPS2,
		fracinv_optim,
			MINFI, &fracinv, fracconst, optfi, &fierr,
		grate_optim,
			MINGE, &Geta, MAXGE, optge, &geerr);

}
